@@ -18,7 +18,9 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
 
 #include <string.h>
 #include <archive_entry.h>
@@ -33,10 +35,10 @@ enum {
 struct _GXPSArchive {
 	GObject parent;
 
-	gboolean  initialized;
-	GError   *init_error;
-	GFile    *filename;
-	GList    *entries;
+	gboolean    initialized;
+	GError     *init_error;
+	GFile      *filename;
+	GHashTable *entries;
 };
 
 struct _GXPSArchiveClass {
@@ -70,7 +72,7 @@ _archive_open (struct archive *archive,
 	return (zip->error) ? ARCHIVE_FATAL : ARCHIVE_OK;
 }
 
-static ssize_t
+static __LA_SSIZE_T
 _archive_read (struct archive *archive,
 	       void           *data,
 	       const void    **buffer)
@@ -87,10 +89,10 @@ _archive_read (struct archive *archive,
 	return read_bytes;
 }
 
-static off_t
+static __LA_INT64_T
 _archive_skip (struct archive *archive,
 	       void           *data,
-	       off_t           request)
+	       __LA_INT64_T    request)
 {
 	ZipArchive *zip = (ZipArchive *)data;
 
@@ -117,9 +119,7 @@ _archive_close (struct archive *archive,
 {
 	ZipArchive *zip = (ZipArchive *)data;
 
-	if (zip->stream)
-		g_object_unref (zip->stream);
-	zip->stream = NULL;
+	g_clear_object (&zip->stream);
 
 	return ARCHIVE_OK;
 }
@@ -165,7 +165,12 @@ gxps_zip_archive_iter_next (ZipArchive            *zip,
 static void
 gxps_zip_archive_destroy (ZipArchive *zip)
 {
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+	/* This is a deprecated synonym for archive_read_free() in libarchive
+	 * 3.0; but is not deprecated in libarchive 2.0, which we continue to
+	 * support. */
 	archive_read_finish (zip->archive);
+G_GNUC_END_IGNORE_DEPRECATIONS
 	g_slice_free (ZipArchive, zip);
 }
 
@@ -174,26 +179,37 @@ gxps_archive_finalize (GObject *object)
 {
 	GXPSArchive *archive = GXPS_ARCHIVE (object);
 
-	if (archive->entries) {
-		g_list_foreach (archive->entries, (GFunc)g_free, NULL);
-		g_list_free (archive->entries);
-		archive->entries = NULL;
-	}
-
-	if (archive->filename) {
-		g_object_unref (archive->filename);
-		archive->filename = NULL;
-	}
-
+	g_clear_pointer (&archive->entries, g_hash_table_unref);
+	g_clear_object (&archive->filename);
 	g_clear_error (&archive->init_error);
 
 	G_OBJECT_CLASS (gxps_archive_parent_class)->finalize (object);
 }
 
+static guint
+caseless_hash (gconstpointer v)
+{
+	gchar *lower;
+	guint ret;
+
+	lower = g_ascii_strdown (v, -1);
+	ret = g_str_hash (lower);
+	g_free (lower);
+
+	return ret;
+}
+
+static gboolean
+caseless_equal (gconstpointer v1,
+                gconstpointer v2)
+{
+	return g_ascii_strcasecmp (v1, v2) == 0;
+}
+
 static void
 gxps_archive_init (GXPSArchive *archive)
 {
-
+	archive->entries = g_hash_table_new_full (caseless_hash, caseless_equal, g_free, NULL);
 }
 
 static void
@@ -264,8 +280,7 @@ gxps_archive_initable_init (GInitable     *initable,
 
         while (gxps_zip_archive_iter_next (zip, &entry)) {
                 /* FIXME: We can ignore directories here */
-                archive->entries = g_list_prepend (archive->entries,
-                                                   g_strdup (archive_entry_pathname (entry)));
+                g_hash_table_add (archive->entries, g_strdup (archive_entry_pathname (entry)));
                 archive_read_data_skip (zip->archive);
         }
 
@@ -290,21 +305,17 @@ gxps_archive_new (GFile   *filename,
 			       NULL);
 }
 
-static inline gboolean
-archive_has_entry (GXPSArchive *archive,
-                   const gchar *path)
-{
-        return g_list_find_custom (archive->entries, path, (GCompareFunc)g_ascii_strcasecmp) != NULL;
-}
-
 gboolean
 gxps_archive_has_entry (GXPSArchive *archive,
 			const gchar *path)
 {
-	if (path && path[0] == '/')
+	if (path == NULL)
+		return FALSE;
+
+	if (path[0] == '/')
 		path++;
 
-	return archive_has_entry (archive, path);
+	return g_hash_table_contains (archive->entries, path);
 }
 
 /* GXPSArchiveInputStream */
@@ -333,14 +344,17 @@ gxps_archive_open (GXPSArchive *archive,
 		   const gchar *path)
 {
 	GXPSArchiveInputStream *stream;
-        gchar                  *first_piece_path = NULL;
+	gchar                  *first_piece_path = NULL;
 
-	if (path && path[0] == '/')
+	if (path == NULL)
+		return NULL;
+
+	if (path[0] == '/')
 		path++;
 
-	if (!archive_has_entry (archive, path)) {
-                first_piece_path = g_build_filename (path, "[0].piece", NULL);
-                if (!archive_has_entry (archive, first_piece_path)) {
+	if (!g_hash_table_contains (archive->entries, path)) {
+                first_piece_path = g_build_path ("/", path, "[0].piece", NULL);
+                if (!g_hash_table_contains (archive->entries, first_piece_path)) {
                         g_free (first_piece_path);
 
                         return NULL;
@@ -407,10 +421,11 @@ gxps_archive_read_entry (GXPSArchive *archive,
 			*bytes_read += bytes;
 		} while (bytes > 0);
 
+		g_object_unref (stream);
+
 		if (*bytes_read == 0) {
 			/* TODO: Error */
 			g_free (*buffer);
-			g_object_unref (stream);
 			return FALSE;
 		}
 
@@ -509,10 +524,7 @@ gxps_archive_input_stream_close (GInputStream  *stream,
 	if (g_cancellable_set_error_if_cancelled (cancellable, error))
 		return FALSE;
 
-	if (istream->zip) {
-		gxps_zip_archive_destroy (istream->zip);
-		istream->zip = NULL;
-	}
+	g_clear_pointer (&istream->zip, gxps_zip_archive_destroy);
 
 	return TRUE;
 }
@@ -522,10 +534,7 @@ gxps_archive_input_stream_finalize (GObject *object)
 {
 	GXPSArchiveInputStream *stream = GXPS_ARCHIVE_INPUT_STREAM (object);
 
-	if (stream->zip) {
-		gxps_zip_archive_destroy (stream->zip);
-		stream->zip = NULL;
-	}
+	g_clear_pointer (&stream->zip, gxps_zip_archive_destroy);
 
 	G_OBJECT_CLASS (gxps_archive_input_stream_parent_class)->finalize (object);
 }
